
/*
 * Copyright (C) 2009 Tudor Marian <tudor.marian@cornell.edu>
 */

#define __FAVOR_BSD /* for BSD style (struct udphdr) */
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <net/ethernet.h>
#include <netinet/ip.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/udp.h>
#include <arpa/inet.h>

#include "list.h"

#ifndef _COMMON64B66B_H_
#define _COMMON64B66B_H_

#define ETH_PREAMBLE_LEN (6+1)
#define ETH_HDR_LEN (2*6+2)
#define ETH_CRC_LEN 4
#define ETH_PREAMBLE_HDR_LEN (ETH_PREAMBLE_LEN + ETH_HDR_LEN)
#define ETH_PREAMBLE_HDR_CRC_LEN (ETH_PREAMBLE_HDR_LEN + ETH_CRC_LEN)

#define	MIN_MTU 46
#define DEFAULT_ETH_MTU 1500
#define JUMBO_ETH_MTU 9000
#define MAX_MTU JUMBO_ETH_MTU
#define MAX_ETH_FRAME	(ETH_PREAMBLE_HDR_CRC_LEN + MAX_MTU)
#define MAX_LEN_INT		(10)
#define MAX_IDLE_LEN	(1+MAX_LEN_INT+2) /* int + sign + new line (\n or \r\n) */

#define IP_HDR_LEN 20
#define UDP_HDR_LEN 8
#ifndef IPPROTO_UDP
#define IPPROTO_UDP 17
#endif
#define ETHER_TYPE_IPV4 0x800

#define NIPQUAD(addr) \
         ((unsigned char *)&addr)[0], \
         ((unsigned char *)&addr)[1], \
         ((unsigned char *)&addr)[2], \
         ((unsigned char *)&addr)[3]
#define NIPQUAD_FMT "%u.%u.%u.%u"

#define MACHEX(addr) \
((unsigned char *)&addr)[0], \
((unsigned char *)&addr)[1], \
((unsigned char *)&addr)[2], \
((unsigned char *)&addr)[3], \
((unsigned char *)&addr)[4], \
((unsigned char *)&addr)[5]
#define MACHEX_FMT "%02x:%02x:%02x:%02x:%02x:%02x"

const uint32_t INIT_CRC=0xffffffff;
const uint32_t CRCPOLY_LE=0xedb88320;
const uint32_t CRCPOLY_BE=0x04c11db7;

/* Iperf first header */
typedef struct UDP_datagram {
#ifdef HAVE_INT32_T
	int32_t id;
	u_int32_t tv_sec;
	u_int32_t tv_usec;
#else
	signed   int id      : 32;
	unsigned int tv_sec  : 32;
	unsigned int tv_usec : 32;
#endif
} UDP_datagram;

/* Iperf second header */
typedef struct client_hdr {
#ifdef HAVE_INT32_T
   /*
    * flags is a bitmap for different options
    * the most significant bits are for determining
    * which information is available. So 1.7 uses
    * 0x80000000 and the next time information is added
    * the 1.7 bit will be set and 0x40000000 will be
    * set signifying additional information. If no
    * information bits are set then the header is ignored.
    * The lowest order diferentiates between dualtest and
    * tradeoff modes, wheither the speaker needs to start
    * immediately or after the audience finishes.
    */
   int32_t flags;
   int32_t numThreads;
   int32_t mPort;
   int32_t bufferlen;
   int32_t mWinBand;
   int32_t mAmount;
#else
   signed int flags      : 32;
   signed int numThreads : 32;
   signed int mPort      : 32;
   signed int bufferlen  : 32;
   signed int mWinBand   : 32;
   signed int mAmount    : 32;
#endif
} client_hdr;

struct frame_bytes
{
	char *buf;
	size_t len, capacity;
};

struct idle_gap
{
	struct list_head list;
	int gap;				/* if gap is negative, fb is set */
	struct frame_bytes fb;
};

struct synced_frame
{
	unsigned char prefix;
	uint64_t payload;
};

uint32_t crc32_le(uint32_t crc, unsigned char *bytes, size_t len)
{
	int i;
	while (len--)
	{
		crc ^= *bytes++;
		for (i = 0; i < 8; i++)
			crc = (crc >> 1) ^ ((crc & 1) ? CRCPOLY_LE : 0);
	}
	return crc;
}

uint32_t crc32_be(uint32_t crc, unsigned char *bytes, size_t len)
{
	int i;
	while (len--)
	{
		crc ^= *bytes++ << 24;
		for (i = 0; i < 8; i++)
			crc = (crc << 1) ^ ((crc & 0x80000000) ? CRCPOLY_BE : 0);
	}
	return crc;
}

uint16_t csum(void *buf, int len)
{
	uint32_t sum = 0;
	unsigned char *cbuf = buf;
	
	while (len > 1)
	{
		sum += (cbuf[0] << 8) | cbuf[1];
		cbuf += 2;
		len -= 2;
	}
	if (len > 0) /* left-over byte, zero padded */
		sum += (cbuf[0] << 8);
	
	/* fold */
	while (sum >> 16)
		sum = (sum & 0xffff) + (sum >> 16);
	return ((uint16_t) (~sum));
}

uint16_t udp_csum(struct udphdr *uh, struct ip *iph)
{
	uint16_t len = ntohs(uh->uh_ulen);
	unsigned char *cbuf = (void *) uh;
	uint64_t sum = 0;
	
	/* the pseudoheader */
	sum += ntohl(iph->ip_src.s_addr);
	sum += ntohl(iph->ip_dst.s_addr);
	sum += (IPPROTO_UDP << 16) | len;

	while (len > 1)
	{
		sum += (cbuf[0] << 8) | cbuf[1];
		cbuf += 2;
		len -= 2;
	}
	if (len > 0) /* left-over byte, zero padded */
		sum += (cbuf[0] << 8);
	
	/* fold */
	while (sum >> 32)
		sum = (sum & 0xffffffff) + (sum >> 32);
	while (sum >> 16)
		sum = (sum & 0xffff) + (sum >> 16);
	sum = ((~sum) & 0xffff);
	if (sum == 0)
		sum = 0xffff;
	return (uint16_t) sum;
}


#define OUT_ERR(e, label)		do { err = (e); goto label; } while(0)

/* returns the size of the list on success, negative number on failure */
int bitstream_from_file(const char *path, struct synced_frame **pframes)
{
	FILE *f = NULL;
	static char buf[MAX_ETH_FRAME], *p;
	int at_frame = 0, frames_cnt = 1024, err;
	struct synced_frame *frames, *tmp;
	
	if (!(f = fopen(path, "r")))
		return -errno;

	if (!(frames = malloc(frames_cnt * sizeof(struct synced_frame))))
		OUT_ERR(-errno, out);
	
	while (fgets(buf, MAX_ETH_FRAME, f))
	{
		long long payload;
		/* skip over leading spaces */
		for (p = buf; *p != '\0' && (*p == ' ' || *p == '\t'); p++)
			;
		/* ignore comments, first, last, and empty line */
		if (*p == '#' || *p == '\n' || *p == ']' || strstr(p, "hbitstream = ["))
			continue;
		if (at_frame == frames_cnt)
		{
			frames_cnt += frames_cnt / 2;
			
			if (!(tmp = malloc(frames_cnt * sizeof(struct synced_frame))))
				OUT_ERR(-errno, out);
			
			memcpy(tmp, frames, at_frame * sizeof(struct synced_frame));
			free(frames);
			frames = tmp;
		}
		sscanf(p, "(%hhx, %llx)", &frames[at_frame].prefix, &payload);
		frames[at_frame].payload = payload;
		at_frame++;
	}
	err = at_frame;
	
	out:
	fclose(f);
	*pframes = frames;
	return err;
}

/* Returns number of packets on success, negative number on failure.
 * If @constant_gap > 0, then each packet except for the last one is
 * followed in the list by a gap of that value.
 */
int packet_list_from_file(const char *path, struct list_head *list, 
	int interleaved_constant_gap)
{
	FILE *f = NULL;
	static char *buf = NULL, *sep_buf, *piece;
	int err, packetcnt = 0;
	
	if (!(f = fopen(path, "r")))
		return -errno;
	
	/* strtok / strsep don't like automatic bufs :( */
	if (!(buf = malloc(MAX_ETH_FRAME)))
		OUT_ERR(-ENOMEM, out);
	
	for (packetcnt = 0; fgets(buf, MAX_ETH_FRAME, f); packetcnt++)
	{
		size_t len_in_bytes = strlen(buf);
		struct idle_gap *g;
				
		if (buf[len_in_bytes] == '\n')
		{
			buf[len_in_bytes] = '\0';
			len_in_bytes -= 1;
		}
		/* each byte is represented as 2 chars, and they are comma delimited */
		len_in_bytes = ((len_in_bytes-2) / 3) + 1;
		
		if (!(g = malloc(sizeof(*g))))
			OUT_ERR(-ENOMEM, out);
		
		if (!(g->fb.buf = malloc(len_in_bytes)))
			OUT_ERR(-ENOMEM, out);

		g->fb.capacity = len_in_bytes;
		g->gap = -1; /* this is not a gap */
		
		sep_buf = buf;
		while ((piece = strsep(&sep_buf, ",")) != NULL)
			sscanf(piece, "%hhx", &g->fb.buf[g->fb.len++]);
		
		list_add_tail(&g->list, list);
		if (interleaved_constant_gap > 0)
		{
			struct idle_gap *const_gap = malloc(sizeof(*const_gap));
			if (!const_gap)
				OUT_ERR(-ENOMEM, out);
			const_gap->gap = interleaved_constant_gap;
			list_add_tail(&const_gap->list, list);
		}
	}
	if ( (interleaved_constant_gap > 0) && (!list_empty(list)) )
	{
		struct idle_gap *last_idle_gap = 
			list_entry(list->prev, struct idle_gap, list);
		
		if (last_idle_gap->gap <= 0)
			OUT_ERR(-EINVAL, out);
		list_del(&last_idle_gap->list);
	}
	err = packetcnt;
	out:
	if (buf)
		free(buf);
	fclose(f);
	return err;
}

int idle_list_from_file(const char *path, struct list_head *list)
{
	FILE *f = NULL;
	static char *buf = NULL;
	int err, idlecnt = 0;
	
	if (!(f = fopen(path, "r")))
		return -errno;
	
	/* strtok / strsep don't like automatic bufs :( */
	if (!(buf = malloc(MAX_IDLE_LEN)))
		OUT_ERR(-ENOMEM, out);
	
	for (idlecnt = 0; fgets(buf, MAX_IDLE_LEN, f); idlecnt++)
	{
		size_t len_in_bytes = strlen(buf);
		struct idle_gap *g;
				
		if (buf[len_in_bytes] == '\n')
		{
			buf[len_in_bytes] = '\0';
			len_in_bytes -= 1;
		}
		
		if (!(g = malloc(sizeof(*g))))
			OUT_ERR(-ENOMEM, out);
		
		g->gap = atoi(buf); /* if negative, this is not a gap */
		if (g->gap <= 0)
			OUT_ERR(-EINVAL, out);
		
		list_add_tail(&g->list, list);
	}
	err = idlecnt;
	out:
	if (buf)
		free(buf);
	fclose(f);
	return err;
}

#endif /* _COMMON64B66B_H_ */

