
/*
 * Copyright (C) 2009 Tudor Marian <tudor.marian@cornell.edu>
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "list.h"
#include "scramblercommon.h"
#include "decodecommon.h"

//#define PRINT_CTRL_CHAR

/* TODO: check for capacity overflow in (struct frame_bytes) buf update */
#define PRINT_BUF_MAX	MAX_MTU
static char print_buf[PRINT_BUF_MAX];

void print_control(char *msg)
{
#ifdef PRINT_CTRL_CHAR
	fprintf(stderr, "%s\n", msg);
#endif /* PRINT_CTRL_CHAR */
}

struct decode_param
{
	struct frame_bytes packet;
	uint64_t bit_position, frame_start, frame_end;
	unsigned long idles;
};

static int parse_control_codes(int from, int to, uint64_t codeword)
{
	unsigned char ctrl_code;
	int i, err_or_reserved = 0;
	
	for (i = 7-from; i > (7-1-to); i--)
	{
		ctrl_code = (codeword >> (i * 7)) & 0x7f;
		if (ctrl_code != 0x0)
			err_or_reserved = 1;		
	}
	return err_or_reserved;
}

static void parse_bytes(int from, int to, uint64_t codeword, char *dest)
{
	int i;
	for (i = 7-from; i > (7-1-to); i--)
	{
		*dest = (codeword >> (i * 8)) & 0xff;
		dest++;
	}
}

void do_decode(enum decode_states *state, uint64_t *descrambler_state, 
			  struct synced_frame *frame, struct decode_param *dp, 
			  accept_fn accept)
{
	unsigned char type;
	uint64_t codeword;
	
	if ((frame->prefix != 0x2) && (frame->prefix != 0x1))
	{
		fprintf(stderr, "Invalid sync preamble, skipping\n");
		*state = start;
		dp->bit_position += 66;
		dp->frame_start = dp->frame_end = 0;
		return;
	}
	
	switch(*state)
	{
		case error:
		{
			fprintf(stderr, "%s\n", dp->packet.buf);
			dp->bit_position += 66;
			dp->frame_start = dp->frame_end = 0;
			*state = start;
			break;
		}
		case start:
		{
			if (frame->prefix == 0x1)
			{
				snprintf(dp->packet.buf, dp->packet.capacity-1, 
					"Started with data codeword");
				*state = error;
			}
			else
			{
#ifdef BITWISE_SCRAMBLER
				codeword = scrambler(descrambler_state, frame->payload, descramble);
#else
				codeword = descrambler64(descrambler_state, frame->payload);
#endif
				type = (codeword >> (64-8)) & 0xff;
				
				switch (type)
				{
					case 0x1e:
					{
						if (parse_control_codes(0, 7, codeword))
						{
							snprintf(dp->packet.buf, dp->packet.capacity-1, 
								"Error or reserved");
							*state = error;
						}
						else
						{
							snprintf(print_buf, PRINT_BUF_MAX, 
								"zzzz (%d idle characters)", 8);
							print_control(print_buf);
							dp->packet.len = 0;
							dp->bit_position += 66;
							dp->idles += 8;
							*state = start;
						}
						break;
					}
					case 0x33:
					{
						if (parse_control_codes(0, 3, codeword))
						{
							snprintf(dp->packet.buf, dp->packet.capacity-1, 
									 "Error or reserved");
							*state = error;
						}
						else
						{
							snprintf(print_buf, PRINT_BUF_MAX, 
									 "zzzz (%d idle characters)", 4);
							print_control(print_buf);
							parse_bytes(5, 7, codeword, 
										&dp->packet.buf[dp->packet.len]);
							dp->packet.len += (7-5+1);
							dp->frame_start = dp->bit_position+(66-3*8);
							dp->bit_position += 66;
							dp->idles += (8-3);
							*state = consume_frame;
						}
						break;
					}
					case 0x78:
					{
						parse_bytes(1, 7, codeword, &dp->packet.buf[dp->packet.len]);
						dp->packet.len += (7-1+1);
						dp->frame_start = dp->bit_position+(2+8);
						dp->bit_position += 66;
						dp->idles += 1;
						*state = consume_frame;
						break;
					}
					case 0x66: /* fall thru */
					case 0x55:
					case 0x4b:
					case 0x2d:
						snprintf(dp->packet.buf, dp->packet.capacity-1, 
							"Ordered set error");
						*state = error;
						break;
					default:
						snprintf(dp->packet.buf, dp->packet.capacity-1,
							"Invalid type to begin with (probably out of sync)");
						*state = error;
						break;
				}
			}
			break;
		}
		case consume_frame:
		{
#ifdef BITWISE_SCRAMBLER
			codeword = scrambler(descrambler_state, frame->payload, descramble);
#else
			codeword = descrambler64(descrambler_state, frame->payload);
#endif
			
			if (frame->prefix == 0x1)
			{
				parse_bytes(0, 7, codeword, &dp->packet.buf[dp->packet.len]);
				dp->packet.len += (7-0+1);
				dp->bit_position += 66;
				/* maintain the state */
			}
			else
			{
				int control_codes;
				type = (codeword >> (64-8)) & 0xff;
				
				switch(type)
				{
					case 0x87:
						control_codes = 7;
						break;
					case 0x99:
						control_codes = 6;
						break;
					case 0xaa:
						control_codes = 5;
						break;
					case 0xb4:
						control_codes = 4;
						break;
					case 0xcc:
						control_codes = 3;
						break;
					case 0xd2:
						control_codes = 2;
						break;
					case 0xe1:
						control_codes = 1;
						break;
					case 0xff:
						control_codes = 0;
						break;
					default:
						snprintf(dp->packet.buf, dp->packet.capacity-1, 
							"Invalid type in consume-frame state %x", type);
						*state = error;
						return; /* no need to continue */
				}
				
				if (parse_control_codes(8-control_codes, 7, codeword))
				{
					snprintf(dp->packet.buf, dp->packet.capacity-1, 
							 "Error or reserved");
					*state = error;
				}
				else
				{
					snprintf(print_buf, PRINT_BUF_MAX, 
							 "zzzz (%d idle characters)", control_codes);
					print_control(print_buf);
					parse_bytes(1, 7-control_codes, codeword, 
								&dp->packet.buf[dp->packet.len]);
					dp->packet.len += ((7-control_codes)-1+1);
					dp->frame_end = dp->bit_position+2+8+(7-control_codes)*8;
					dp->idles = control_codes + 1;
					*state = accept_frame;
				}
			}

			break;
		}
		case accept_frame:
		{
			char *msg = print_buf;
			struct iperf_id_and_ip_id_ret *ids = NULL;
			
			accept(&dp->packet, print_buf);
			if (accept != raw_eth_vector_decoded)
			{
				ids = (void *)print_buf;
				msg = (void *)print_buf + sizeof(struct iperf_id_and_ip_id_ret);
			}
			fprintf(stdout, "%s", msg);
			if (ids)
				fprintf(stdout, "Iperf id %d , IP id %d , start bit %lld , "
					"end bit %lld\n", ids->iperf_id, ids->ip_id, 
					(long long) dp->frame_start, (long long) dp->frame_end);
			
			dp->packet.len = 0;
			dp->bit_position += 66;
			dp->frame_start = dp->frame_end = 0;
			*state = start;
			break;
		}
		default:
			/* not reached */
			break;
	}
}

/* returns 0 on success, should collect all packets into a list */
int decode_(enum decode_states state, uint64_t descrambler_state, 
	struct synced_frame *frames, int framecnt, accept_fn accept, 
	struct list_head *idle_gaps)
{
	int err = 0, i = 0;
	struct decode_param dp;
	struct idle_gap *gap;
	
	memset(&dp, 0, sizeof(dp));

	if ((!frames) || (framecnt <= 0))
	{
		fprintf(stderr, "No input.\n");
		err = -1;
		goto done;
	}
	
	if (!(dp.packet.buf = malloc(MAX_MTU)))
	{
		fprintf(stderr, "Unable to allocate MTU size packet pad\n");
		err = -ENOMEM;
		goto done;
	}
	dp.packet.capacity = MAX_MTU;
	
	while (i < framecnt)
	{
		enum decode_states oldstate = state;
		do_decode(&state, &descrambler_state, &frames[i], &dp, accept);
		if ((state != error) && (state != accept_frame))
			i++;
		if ((oldstate == start) && (state == consume_frame))
		{
			if (!(gap = malloc(sizeof(*gap))))
			{
				fprintf(stderr, "Unable to allocate (struct idle_gap)\n");
				err = -ENOMEM;
				goto done;
			}
			gap->gap = dp.idles;
			list_add_tail(&gap->list, idle_gaps);
		}
	}
	if (!(gap = malloc(sizeof(*gap))))
	{
		fprintf(stderr, "Unable to allocate (struct idle_gap)\n");
		err = -ENOMEM;
		goto done;
	}
	gap->gap = dp.idles;
	list_add_tail(&gap->list, idle_gaps);
	
	done:
	if (dp.packet.buf)
		free(dp.packet.buf);
	return err;
}

int decode(struct synced_frame *frames, int framecnt, accept_fn accept, 
	struct list_head *gaps)
{
	return decode_(start, 0xffffffffffffffffLL, frames, framecnt, accept, gaps);
}

int main(int argc, char *argv[])
{
	int i, framecnt, noiperf = 0;
	struct synced_frame *frames;
	struct idle_gap *gap, *dummy;
	accept_fn acceptor = print_iperf_vector_decoded;
	char *bitstream_file = NULL;
	
	DEFINE_LIST_HEAD(idle_gaps);
	
	for (i = 1; i < argc; i++)
	{
		if (strcmp(argv[i], "-o") == 0)
			acceptor = raw_eth_vector_decoded;
		else if (strcmp(argv[i], "-n") == 0)
			noiperf = 1;
		else
			bitstream_file = argv[i];
	}
	
	if (noiperf)
	{
		if (acceptor == raw_eth_vector_decoded)
			fprintf(stderr, "-o flag supersedes and cancels the -n flag\n");
		else
			acceptor = print_eth_vector_decoded;
	}
	
	if (!bitstream_file)
	{
		fprintf(stderr, "Usage %s [-o (serialize packets to stdout)] "\
			" [-n (do not attempt to parse iperf header)] <bitstream-file>\n", 
			argv[0]);
		exit(EXIT_FAILURE);
	}
	
	if ((framecnt = bitstream_from_file(bitstream_file, &frames)) < 0)
	{
		fprintf(stderr, "Error %s\n", strerror(framecnt));
		exit(EXIT_FAILURE);
	}
	
	decode(frames, framecnt, acceptor, &idle_gaps);
	
	fprintf(stderr, "Idle gaps: [");
	list_for_each_entry_safe(gap, dummy, &idle_gaps, list)
	{
		list_del(&gap->list);
		if (list_empty(&idle_gaps))
			fprintf(stderr, "%d]\n", gap->gap);
		else
			fprintf(stderr, "%d, ", gap->gap);
		free(gap);
	}
	
	exit(EXIT_SUCCESS);
}
