
# Copyright (C) 2009 Tudor Marian <tudor.marian@cornell.edu>

import sys
from decodecommon import *
from codegroups8b10b import *

comma_states_max = 6 # actually 7, but the last one is accept
comma_state_accept = 7
comma_states_input_plus = ['0', '0', '1', '1', '1', '1', '1']
comma_states_input_minus = ['1', '1', '0', '0', '0', '0', '0']

def comma_states_automata(bit, state, is_rd_plus):
	if state < 0 or state > comma_states_max:
		raise DecodeException('Invalid comma state %d, terminating' % state)
	if bit != '0' and bit != '1':
		raise DecodeException('Comma parsing bits, bit not 0 or 1')
	if is_rd_plus:
		comma_states_input = comma_states_input_plus
	else:
		comma_states_input = comma_states_input_minus
	if comma_states_input[state] == bit:
		return state+1
	else:
		return 0
#end comma_states_automata

# reads a vanilla bitstream, single bit per line
def read_bitstream(f):
	comma_state_p = 0
	comma_state_m = 0
	vector = []
	tenbits = 0x0
	atbit = 0
	comma_found = False
	for line in f:
		line = line.strip()
		if not line or line.startswith('#'):
			continue
		if not comma_found:
			comma_state_p = comma_states_automata(line, comma_state_p, True)
			if comma_state_p == comma_state_accept:
				comma_found = True
				tenbits = int(reduce(lambda x, y:x+y, comma_states_input_plus), 2)
				atbit = 7
			comma_state_m = comma_states_automata(line,comma_state_m, False)
			if comma_state_m == comma_state_accept:
				comma_found = True
				tenbits = int(reduce(lambda x, y:x+y, comma_states_input_minus), 2)
				atbit = 7
		else:
			bit = int(line)
			assert (bit == 1 or bit == 0)
			tenbits = (tenbits << 1) | bit
			atbit += 1
			# this will truncate at 10 bit boundary
			if atbit == 10:
				vector.append(tenbits)
				tenbits = 0
				atbit = 0
	return vector
#end read_bitstream

# this is deprecated, it only reads 10-bit aligned integers from a test vector
def read_test_vector(f):
	vec = []
	for line in f:
		if not line or line.strip().startswith('#'):
			continue
		pieces = line.strip().split()
		vec.append(int(pieces[3], 2))
	return vec
#end read_test_vector

# vector is an array of bytes---it has already been synchronized w.r.t. `commas'
def do_decode(state, vector, param, codegroups, accept_fn):
	if not vector:
		return None
	
	frame_bytes, bit_position, frame_start, frame_end = param
	
	if state is 'error':
		sys.stderr.write(str(frame_bytes))
		param = ('', bit_position+8, None, None)
		del vector[0] # peell off mutable list instead of passing a copy vector[1:]
		return ('start', vector, param)
	elif state is 'start':
		(name, octet_val, special) = codegroups[vector[0]]
		if not special:
			err = 'Started with data codeword\n'
			param = (err, bit_position, frame_start, frame_end)
			return ('error', vector, param)
		elif config_or_idle(name):
			if len(vector) == 1:
				return None # lookahead has its price
			(name_, octet_val_, special_) = codegroups[vector[1]]
			if config_and_not_idle(name_):
				if len(vector) < 4:
					return None
				config_reg_1 = codegroups[vector[2]]
				config_reg_2 = codegroups[vector[3]]
				sys.stderr.write('Config regs: %s %s\n' % \
					(config_reg_1[0], config_reg_2[0]))
				param = ('', bit_position+4*10, frame_start, frame_end)
				del vector[:4]
				return ('start', vector, param)
			else:
				assert idle1_or_2(name_)
				param = ('', bit_position+2*10, frame_start, frame_end)
				del vector[:2]
				return ('start', vector, param)
		elif carrier_extended(name):
			err = 'Carrier extend\n'
			param = (err, bit_position, frame_start, frame_end)
			return ('error', vector, param)
		elif end_of_packet(name):
			err = 'End of packet before start\n'
			param = (err, bit_position, frame_start, frame_end)
			return ('error', vector, param)
		elif error_propagation(name):
			err = 'Error propagation before start\n'
			param = (err, bit_position, frame_start, frame_end)
			return ('error', vector, param)
		elif start_of_packet(name):
			param = (frame_bytes, bit_position+10, bit_position+10, None)
			del vector[0]
			return ('consume-frame', vector, param)
		else:
			err = 'Invalid type %s before start\n' % name
			param = (err, bit_position, frame_start, frame_end)
			return ('error', vector, param)
	elif state is 'consume-frame':
		(name, octet_val, special) = codegroups[vector[0]]
		if not special:
			param = (frame_bytes+chr(octet_val), \
				bit_position+10, frame_start, frame_end)
			del vector[0]
			return ('consume-frame', vector, param)
		else:
			if end_of_packet(name):
				param = (frame_bytes, bit_position, frame_start, bit_position+10)
				return ('accept-frame', vector, param)
			else:
				err = 'Invalid type %s in consume-frame state' % name
				param = (err, bit_position, frame_start, frame_end)
				return ('error', vector, param)
	elif state is 'accept-frame':
		accept_ret = accept_fn(frame_bytes)
		if accept_ret:
			assert (isinstance(accept_ret, tuple)) and (len(accept_ret) == 4)
			iperf_id, iperf_sec, iperf_usec, msg = accept_ret
			sys.stderr.write(msg)
			sys.stdout.write('Packet id %d, start bit %s, end bit %s\n' % \
				(iperf_id, repr(frame_start), repr(frame_end)))
		param = ('', bit_position+10, None, None)
		del vector[0]
		return ('start', vector, param)
	else:
		raise DecodeException('Invalid state, terminating')
#end do_decode

def decode_(state, vector, codegroups, accept_fn):
	param = ('', 0, None, None)
	if not vector:
		sys.stderr.write('No input.\n')
	else:
		while True:
			ret = do_decode(state, vector, param, codegroups, accept_fn)
			if not ret:
				sys.stderr.write('Consumed all input.\n')
				return
			state, vector, param = ret
#end decode_

def decode(vector, codegroups, accept_fn):
	decode_('start', vector, codegroups, accept_fn)
#end decode

def main():
	if len(sys.argv) != 2:
		sys.stderr.write('Usage %s <test-vector-file>\n' % (sys.argv[0]))
		sys.exit(1)
	f = open(sys.argv[1])
	test_vector = read_bitstream(f)
	f.close()
	
	map = code_groups_map(data_code_groups, special_code_groups)
	decode(test_vector, map, eth_vector_decoded)
#end main

if __name__ == "__main__":
	main()
