
# Copyright (C) 2009 Tudor Marian <tudor.marian@cornell.edu>

import sys,struct,socket

def raw_eth_vector_decoded(vector):
	for c in vector:
		sys.stderr.write('%02x ' % ord(c))
	sys.stderr.write('\n')
#end raw_eth_vector_decoded

states = set(['error', 'start', 'consume-frame', 'accept-frame'])

class DecodeException(Exception):
	def __init__(self, value):
		self.value = value
	def __str__(self):
		return repr(self.value)
#end class DecodeException

ETH_PREAMBLE_LEN = 6+1
ETH_HDR_LEN = 2*6+2
ETH_CRC_LEN = 4
ETH_PREAMBLE_HDR_CRC_LEN = ETH_PREAMBLE_LEN + ETH_HDR_LEN + ETH_CRC_LEN

MIN_MTU = 46
DEFAULT_ETH_MTU = 1500
JUMBO_ETH_MTU = 9000
MAX_MTU = JUMBO_ETH_MTU

IP_HDR_LEN = 20
UDP_HDR_LEN = 8
IPPROTO_UDP = 17
ETHER_TYPE_IPV4 = 0x800

INIT_CRC=0xffffffff
CRCPOLY_LE=0xedb88320
CRCPOLY_BE=0x04c11db7

#precomputed
range8 = range(8)
range4 = range(4)

def crc32_le(crc, bytes):
	for str_byte in bytes:
		byte = ord(str_byte) & 0xff
		crc ^= byte
		for i in range8:
			crc = (crc >> 1) ^ (CRCPOLY_LE if (crc & 0x1 != 0) else 0)
			crc &= INIT_CRC
	return crc
#end crc43_le

def crc32_be(crc, bytes):
	for str_byte in bytes:
		byte = ord(str_byte) & 0xff
		crc ^= (byte << 24)
		for i in range8:
			crc = (crc << 1) ^ (CRCPOLY_BE if ((crc & 0x80000000) != 0) else 0)
			crc &= INIT_CRC
	return crc
#end crc32_be

def udp_csum(udp_header, length, ip_header_addrs, payload):
	csum = bytes0_4 = bytes4_8 = bytes8_12 = bytes12_16 = bytes16_20 = 0L
	for i in range4:
		bytes0_4 |= ord(ip_header_addrs[i]) << (8 * (3-i))
		bytes4_8 |= ord(ip_header_addrs[4+i]) << (8 * (3-i))
		bytes12_16 |= ord(udp_header[i]) << (8 * (3-i))
	bytes8_12 = (IPPROTO_UDP << 16) | (length & 0xffff)
	bytes16_20 = (length & 0xffff) << 16 
	
	csum += bytes0_4 + bytes4_8 + bytes8_12 + bytes12_16 + bytes16_20
	
	for i in range(len(payload) / 2):
		csum += (ord(payload[2*i]) << 8) | ord(payload[2*i+1])
	if len(payload) % 2:
		csum += (ord(payload[-1]) << 8)
	
	# fold twice
	while csum >> 32 != 0:
		csum = (csum & 0xffffffff) + (csum >> 32)
	while csum >> 16 != 0:
		csum = (csum & 0xffff) + (csum >> 16)
	csum = ((~csum) & 0xffff)
	if csum == 0:
		csum = 0xffff
	return csum
#end udp_csum

def ipv4_header_csum(header, header_addrs, options):
	csum = bytes0_4 = bytes4_8 = bytes8_12 = bytes12_16 = bytes16_20 = 0L
	for i in range4:
		bytes0_4 |= ord(header[i]) << (8 * (3-i))
		bytes4_8 |= ord(header[4+i]) << (8 * (3-i))
		bytes12_16 |= ord(header_addrs[i]) << (8 * (3-i))
		bytes16_20 |= ord(header_addrs[4+i]) << (8 * (3-i))
	# the header does not include the checksum
	bytes8_12 |= ( ord(header[8]) << (8 * 3) ) | ( ord(header[9]) << (8 * 2) )
	# use the 64 bit width of python integers
	csum += bytes0_4 + bytes4_8 + bytes8_12 + bytes12_16 + bytes16_20
	# options are multiple of 32-bit words
	assert len(options) % 4 == 0
	for i in range(len(options) / 4):
		options_4_bytes = 0L
		for j in range4:
			options_4_bytes |= ord(options[4*i + j]) << (8 * (3-j))
		csum += options_4_bytes  
	# fold once
	while csum >> 32 != 0:
		csum = (csum & 0xffffffff) + (csum >> 32)
	# fold twice
	while csum >> 16 != 0:
		csum = (csum & 0xffff) + (csum >> 16)
	return ((~csum) & 0xffff)
#end ipv4_header_csum

def print_eth_vector_decoded(vector):
	msg = ''
	# 6 x 0x55 bytes
	for b in vector[:6]:
		if ord(b) != 0x55:
			sys.stderr.write('Invalid preamble bytes\n')
			return None
	# SOF, little endian 10101011
	if ord(vector[6]) != 0xd5:
		sys.stderr.write('Invalid SOF\n')
		return None
	crc_bytes_str = vector[-4:]
	frame = vector[7:-4]
	mac_dst = ':'.join([hex(ord(b))[2:].zfill(2) for b in frame[:6]])
	mac_src = ':'.join([hex(ord(b))[2:].zfill(2) for b in frame[6:12]])
	ether_type = (ord(frame[12]) << 8) | (ord(frame[13]))
	assert (ether_type == ETHER_TYPE_IPV4) # must be IPv4
	tabs = 0
	msg += '%sMAC:\n' % (tabs * '\t')
	tabs += 1
	msg += '%sDst addr: %s, Src addr: %s\n\tEtherType: 0x%X\n' % \
		(tabs * '\t', mac_dst, mac_src, ether_type)
	crc = 0
	# make sure I reverse the crc bytes (we use little endian)
	for crc_byte_str in crc_bytes_str[::-1]:
		crc_byte = ord(crc_byte_str)
		crc = (crc << 8) | crc_byte
	crc_le = crc32_le(INIT_CRC, frame) ^ INIT_CRC
	crc_check = 'correct' if (crc == crc_le) else 'INCORRECT!'
	msg += '%sCRC: 0x%08x (%s)\n' % (tabs * '\t', crc, crc_check)
	
	# IP
	payload = frame[14:]
	ip_version = ord(payload[0]) >> 4
	ip_hdr_len = ( ord(payload[0]) & 0xf ) * 4
	ip_tos = ord(payload[1])
	ip_tot_len = (ord(payload[2]) << 8) | ord(payload[3])
	ip_id = (ord(payload[4]) << 8) | ord(payload[5])
	ip_frag_off = ( (ord(payload[6]) << 8) | ord(payload[7]) ) & 0x1fff
	ip_flags = ord(payload[6]) >> 5
	ip_ttl = ord(payload[8])
	ip_protocol = ord(payload[9])
	ip_hdr_check = (ord(payload[10]) << 8) | ord(payload[11])
	assert (ip_protocol == IPPROTO_UDP) # UDP or bust
	ip_src_addr = '.'.join([repr(ord(b)) for b in payload[12:16]])
	ip_dst_addr = '.'.join([repr(ord(b)) for b in payload[16:20]])
	ip_options_len = ip_hdr_len - IP_HDR_LEN
	
	ip_hdr_addrs = payload[12:20]
	ip_hdr_csum = ipv4_header_csum(payload, ip_hdr_addrs, payload[20:ip_hdr_len])
	ip_hdr_csum_check = 'correct' if (ip_hdr_check==ip_hdr_csum) else 'INCORRECT!'
	
	msg += '\n%sIP:\n' % (tabs * '\t')
	tabs += 1
	msg += '%sVersion: %d, Hdr len %d, IP TOS: %d, IP total length: %d\n' % \
		(tabs * '\t', ip_version, ip_hdr_len, ip_tos, ip_tot_len)
	msg += '%sID: %d, Flags: %x, Fragment offset: %d, TTL: %d\n' % \
		(tabs * '\t', ip_id, ip_flags, ip_frag_off, ip_ttl)
	msg += '%sProtocol: %d, Hdr checksum: 0x%04x (%s)\n' % \
		(tabs * '\t', ip_protocol, ip_hdr_check, ip_hdr_csum_check)
	msg += '%sSrc addr: %s, Dst addr: %s\n' % \
		(tabs * '\t', ip_src_addr, ip_dst_addr)
	if ip_options_len > 0:
		msg += '%s Options: %d bytes worth\n' % (tabs * '\t', ip_options_len)
	
	# UDP
	payload = payload[ip_hdr_len:]
	udp_source = (ord(payload[0]) << 8) | ord(payload[1])
	udp_dest = (ord(payload[2]) << 8) | ord(payload[3])
	udp_len = (ord(payload[4]) << 8) | ord(payload[5])
	udp_check = (ord(payload[6]) << 8) | ord(payload[7])
	udp_payload = payload[UDP_HDR_LEN:]
	
	udp_check_verif = udp_csum(payload, udp_len, ip_hdr_addrs, udp_payload)
	udp_csum_check = 'correct' if (udp_check==udp_check_verif) else 'INCORRECT!'
	
	msg += '\n%sUDP:\n' % (tabs * '\t')
	tabs += 1
	msg += '%sSrc port: %d, Dst port: %d, Length: %d\n' % \
		(tabs * '\t', udp_source, udp_dest, udp_len)
	msg += '%sChecksum: 0x%04x (%s)\n' % (tabs * '\t', udp_check, udp_csum_check)
	return (msg, ip_id, udp_payload, tabs)
#end print_eth_vector_decoded

def print_iperf_vector_decoded(vector):
	msg, ip_id, udp_payload, tabs = print_eth_vector_decoded(vector)
	# Iperf UDP payload, typedef struct UDP_datagram { ... } UDP_datagram;
	iperf_id, iperf_sec, iperf_usec = struct.unpack('III', udp_payload[:4*3])
	iperf_id = socket.ntohl(iperf_id)
	iperf_sec = socket.ntohl(iperf_sec)
	iperf_usec = socket.ntohl(iperf_usec)
	
	msg += '\n%sIperf:\n' % (tabs * '\t')
	tabs += 1
	msg += '%sID: %d\n%sTimestamp (sec, usec): %u %u\n' % \
		(tabs * '\t', iperf_id, tabs * '\t', iperf_sec, iperf_usec)
	return (msg, iperf_id, ip_id)
#end print_iperf_vector_decoded
