
/*
 * Copyright (C) 2009 Tudor Marian <tudor.marian@cornell.edu>
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "scramblercommon.h"
#include "common64b66b.h"

#define INTER_FRAME_GAP	12 /* 96 bit times inter-frame gap */

const uint64_t idle_frame = 0x1e00000000000000LL;
const uint64_t scrambler_start_state = 0xffffffffffffffffLL;

const char *print_buf[MAX_MTU];

#define PRINT_FMT		"(0x%1x, 0x%016llx),\\\n"
#define PRINT_FMT_LAST	"(0x%1x, 0x%016llx)]\n"

/* returns 0 on success, fd has to be an open file descriptor */
int do_encode_with_fixed_idle_gap(struct list_head *packets, 
	uint64_t scrambler_state, int leading_idles, int trailing_idles, FILE *fd)
{
	struct idle_gap *g, *tmp;
	uint64_t scrambled_frame;
	int i;
	
	/* Idles are given in bytes, however, we divide them up into 64bit frames */
	leading_idles /= 8;
	for (i = 0; i < leading_idles; i++)
	{
#ifdef BITWISE_SCRAMBLER
		scrambled_frame = scrambler(&scrambler_state, idle_frame, scramble);
#else
		scrambled_frame = scrambler64(&scrambler_state, idle_frame);
#endif
		if (fprintf(fd, PRINT_FMT, 0x2, (long long)scrambled_frame) < 0)
			return -errno;
	}
	
	list_for_each_entry_safe(g, tmp, packets, list)
	{
		list_del(&g->list);
	
		if (g->gap > 0) /* it's a gap */
		{
			if (g->gap < INTER_FRAME_GAP)
			{
				fprintf(stderr, "Gap > interframe gap, adjusting...\n");
				g->gap = INTER_FRAME_GAP;
			}
			for (i = 0; i < g->gap / 8; i++)
			{
#ifdef BITWISE_SCRAMBLER
				scrambled_frame = scrambler(&scrambler_state, idle_frame, scramble);
#else
				scrambled_frame = scrambler64(&scrambler_state, idle_frame);
#endif
				if (fprintf(fd, PRINT_FMT, 0x2, (long long)scrambled_frame) < 0)
					return -errno;
			}
		}
		else /* it's a packet */
		{
			int len = g->fb.len, start_frame_bytes, tail_frame_bytes, i, j;
			unsigned char *bytes = (unsigned char *) g->fb.buf;
			uint64_t start_frame, frame;
			
			/* issue packet */
			if ( ((len - 3) % 8) < ((len - 7) % 8) )
			{
				start_frame = 0x3300000000LL;
				start_frame_bytes = 3;
			}
			else
			{
				start_frame = 0x78;
				start_frame_bytes = 7;
			}
			
			/* /S/ */
			for (i = 0; i < start_frame_bytes; i++)
				start_frame = (start_frame << 8) | bytes[i];
#ifdef BITWISE_SCRAMBLER
			scrambled_frame = scrambler(&scrambler_state, start_frame, scramble);
#else
			scrambled_frame = scrambler64(&scrambler_state, start_frame);
#endif
			if (fprintf(fd, PRINT_FMT, 0x2, (long long)scrambled_frame)<0)
				return -errno;
			
			bytes += start_frame_bytes;
			len -= start_frame_bytes;
			tail_frame_bytes = len % 8;
		
			/* fill in the bulk of the packets */
			for (i = 0; i < len / 8; i++)
			{
				unsigned char *slice = &bytes[i*8];
				
				frame = 0;
				
				for (j = 0; j < 8; j++)
					frame = (frame << 8) | slice[j];
#ifdef BITWISE_SCRAMBLER
				scrambled_frame = scrambler(&scrambler_state, frame, scramble);
#else
				scrambled_frame = scrambler64(&scrambler_state, frame);
#endif
				if (fprintf(fd, PRINT_FMT, 0x1, (long long)scrambled_frame) < 0)
					return -errno;
			}
			
			/* /T/ */
			switch (tail_frame_bytes)
			{
				case 0:
					frame = 0x87;
					break;
				case 1:
					frame = 0x99;
					break;
				case 2:
					frame = 0xaa;
					break;
				case 3:
					frame = 0xb4;
					break;
				case 4:
					frame = 0xcc;
					break;
				case 5:
					frame = 0xd2;
					break;
				case 6:
					frame = 0xe1;
					break;
				default:
					if (tail_frame_bytes != 7)
					{
						fprintf(stderr, "/T/ assertion error\n");
					exit(EXIT_FAILURE);
				}
				frame = 0xff;
				break;
			}
		
			/* add the remaining data bytes, if any */
			for (i = (len / 8) * 8; i < len; i++)
				frame = (frame << 8) | bytes[i];
			/* add the trailing idle control codes */
			for (i = 0; i < (7-tail_frame_bytes); i++)
				frame = (frame << 8) | 0x0;
#ifdef BITWISE_SCRAMBLER
			scrambled_frame = scrambler(&scrambler_state, frame, scramble);
#else
			scrambled_frame = scrambler64(&scrambler_state, frame);
#endif
			if (fprintf(fd, PRINT_FMT, 0x2, (long long)scrambled_frame) < 0)
				return -errno;
			
			free(g->fb.buf);
		}
		free(g);
	}
	
	trailing_idles /= 8; /* normalize to 64 bit frame */
	for (i = 0; i < trailing_idles; i++)
	{
#ifdef BITWISE_SCRAMBLER
		scrambled_frame = scrambler(&scrambler_state, idle_frame, scramble);
#else
		scrambled_frame = scrambler64(&scrambler_state, idle_frame);
#endif
		if (i == trailing_idles - 1)
		{
			if (fprintf(fd, PRINT_FMT_LAST, 0x2, (long long)scrambled_frame) < 0)
				return -errno;
		}
		else
		{
			if (fprintf(fd, PRINT_FMT, 0x2, (long long)scrambled_frame) < 0)
				return -errno;
		}
	}
	
	return 0;
}

static inline int encode_with_fixed_idle_gap(struct list_head *packets, 
	int leading_idles, int trailing_idles, FILE *filep)
{
	return do_encode_with_fixed_idle_gap(packets, scrambler_start_state, 
		leading_idles, trailing_idles, filep);
}

int main(int argc, char *argv[])
{
	FILE *f = stdout;
	char *packet_list_file = NULL;
	char *idle_list_file = NULL;
	int packetcnt, err, i;
	int idle_gap = 0, leading_idles = 0, trailing_idles = 0;
	
	DEFINE_LIST_HEAD(packet_list);
	
	for (i = 1; i < argc; i++)
	{
		if (strcmp(argv[i], "--constant-idle-gap") == 0)
			idle_gap = atoi(argv[++i]);
		else if (strcmp(argv[i], "--list-of-idles-file") == 0)
			idle_list_file = argv[++i];
		else if (strcmp(argv[i], "--leading-idles") == 0)
			leading_idles = atoi(argv[++i]);
		else if (strcmp(argv[i], "--trailing-idles") == 0)
			trailing_idles = atoi(argv[++i]);
		else
			packet_list_file = argv[i];
	}
	
	if (!packet_list_file)
	{
		fprintf(stderr, "Usage %s [--constant-idle-gap gap] " \
			"[--list-of-idles-file <file>] [--leading-idles idles] " \
			"[--trailing-idles idles] <packet list file>\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	
	if (idle_gap > 0 && idle_list_file)
	{
		fprintf(stderr, "Warning: --list-of-idles-file trumps --constant-idle-gap");
		idle_gap = -1;
	}
	
	if (!idle_list_file && idle_gap < INTER_FRAME_GAP)
	{
		fprintf(stderr, "Will issue at least %d idles due to inter-frame gap "
			"constraints\n", INTER_FRAME_GAP);
		idle_gap = INTER_FRAME_GAP;
	}
	
	if (leading_idles < 8)
	{
		fprintf(stderr, "Will issue at least 8 leading idles for scrambler sync\n");
		leading_idles = 8;
	}
	
	if (trailing_idles < 8)
	{
		fprintf(stderr, "Will issue at least 8 trailing idles for bitstream " \
			"formatting\n");
		trailing_idles = 8;
	}
	
	if ((packetcnt = packet_list_from_file(packet_list_file, &packet_list, 
		idle_gap)) < 0)
	{
		fprintf(stderr, "Error %s\n", strerror(-packetcnt));
		exit(EXIT_FAILURE);
	}
	
	if (idle_list_file)
	{
		int idlecnt = 0;
		struct idle_gap *next_packet;
		struct idle_gap *next_idle;
		
		DEFINE_LIST_HEAD(idle_list);
		DEFINE_LIST_HEAD(combined_list);
		
		if ((idlecnt = idle_list_from_file(idle_list_file, &idle_list)) < 0)
		{
			fprintf(stderr, "Error %s\n", strerror(-idlecnt));
			exit(EXIT_FAILURE);
		}
		
		if (idlecnt != packetcnt-1)
		{
			fprintf(stderr, "Error: there should be N-1 idles, where N is " \
				"the number of packets. Instead, there are %d idles and %d " \
				"packets\n", idlecnt, packetcnt);
			exit(EXIT_FAILURE);
		}
		/* merging the idles w/ the packets */
		while (!list_empty(&idle_list))
		{
			next_packet = list_entry(packet_list.next, struct idle_gap, list);
			next_idle = list_entry(idle_list.next, struct idle_gap, list);
			list_del(&next_packet->list);
			list_add_tail(&next_packet->list, &combined_list);
			list_del(&next_idle->list);
			list_add_tail(&next_idle->list, &combined_list);
		}
		next_packet = list_entry(packet_list.next, struct idle_gap, list);
		list_del(&next_packet->list);
		list_add_tail(&next_packet->list, &combined_list);
		
		if (!list_empty(&packet_list))
		{
			fprintf(stderr, "Assertion failure: lists inconsistent\n");
			exit(EXIT_FAILURE);
		}
		list_replace(&combined_list, &packet_list);
	}
	
	fprintf(f, "hbitstream = [\\\n");
	err = encode_with_fixed_idle_gap(&packet_list, leading_idles, trailing_idles, f);
	if (err < 0)
	{
		fprintf(stderr, "Error %s\n", strerror(-err));
		exit(EXIT_FAILURE);
	}
			
	exit(EXIT_SUCCESS);
}
