
# Copyright (C) 2009 Tudor Marian <tudor.marian@cornell.edu>

import sys,math,pickle
from scramblercommon import *
from common64b66b import *

oneKbps = math.pow(10, 3) / 8
oneMbps = math.pow(10, 6) / 8
oneGbps = math.pow(10, 9) / 8
tenGbps = 10 * oneGbps
maxTxRate = tenGbps

INTER_FRAME_GAP=12 # 96 bit times inter-frame gap
idle_frame=0x1e00000000000000
scrambler_start_state=0xffffffffffffffff

# idles are given in number of BYTES, not actual idles, since an idle character
# is 7 bit instead of 8 bit. Block type fields are considered as part of idle
# payload, even if /S/ or /T/ blocks are sent.
def do_encode(vector, scrambler_state, track_excess_idles):
	excess_idles = 0
	bitstream = []
	for packet in vector:
		if not track_excess_idles:
			excess_idles = 0
		if isinstance(packet, int) or isinstance(packet, long):
			# packet encodes the number of idles
			assert (packet >= 0)
			#sys.stderr.write('Idles: %d, excess: %d\n' % (packet, excess_idles))
			if packet < excess_idles:
				excess_idles -= packet
				packet = 0
			else:
				packet -= excess_idles
				excess_idles = 0
				# a pure idle frame = 8x7bit idles + 1x8bit block type = 8bytes
				i = 0
				while i < (packet / 8):
					scrambler_state, scrambled_frame = \
						scrambler(scrambler_state, idle_frame, scramble)
					bitstream.append((0x2, scrambled_frame))
					i += 1
				# don't forget the leftovers
				#if track_excess_idles:
				excess_idles += (packet % 8)
		elif isinstance(packet, str):
			if ((len(packet) - 3) % 8) < ((len(packet) - 7) % 8):
				start_frame = 0x3300000000
				start_frame_bytes = 3
				excess_idles += (8 - 3) # block type + 4 idles + 4 bits
			else:
				start_frame = 0x78
				start_frame_bytes = 7
				excess_idles += 1 # block type
			for i in range(start_frame_bytes):
				start_frame = (start_frame << 8) | ord(packet[i])
			scrambler_state, scrambled_frame = \
				scrambler(scrambler_state, start_frame, scramble)
			bitstream.append((0x2, scrambled_frame))
			
			packet = packet[start_frame_bytes:]
			tail_frame_bytes = len(packet) % 8
			# fill in the regular packets
			for i in range(len(packet) / 8):
				slice = packet[i*8:i*8+8]
				frame = 0x0
				for byte in slice:
					frame = (frame << 8) | ord(byte)
				scrambler_state, scrambled_frame = \
					scrambler(scrambler_state, frame, scramble)
				bitstream.append((0x1, scrambled_frame))
			# the /T/
			if tail_frame_bytes == 0:
				frame = 0x87
			elif tail_frame_bytes == 1:
				frame = 0x99
			elif tail_frame_bytes == 2:
				frame = 0xaa
			elif tail_frame_bytes == 3:
				frame = 0xb4
			elif tail_frame_bytes == 4:
				frame = 0xcc
			elif tail_frame_bytes == 5:
				frame = 0xd2
			elif tail_frame_bytes == 6:
				frame = 0xe1
			else:
				assert (tail_frame_bytes == 7)
				frame = 0xff
			# add the remaining data bytes, if any
			for byte in packet[((len(packet) / 8) * 8):]:
				frame = (frame << 8) | ord(byte)
			# add the trailing idle control codes
			for i in range(7-tail_frame_bytes):
				frame = (frame << 8) | 0x0
			scrambler_state, scrambled_frame = \
				scrambler(scrambler_state, frame, scramble)
			bitstream.append((0x2, scrambled_frame))
			excess_idles += (8 - tail_frame_bytes)
			# if enough excess idles have accumulated, issue them out now
			#if track_excess_idles:
			#	for i in range(excess_idles / 8):
			#		scrambler_state, scrambled_frame, input_bits, output_bits = \
			#		    scrambler(scrambler_state, idle_frame, scramble)
			#		bitstream.append((0x2, scrambled_frame))
			#	excess_idles = excess_idles % 8
		else:
			sys.stderr.write('Invalid packet type\n')
			sys.exit(1)
	return (bitstream, excess_idles)
#end do_encode

def encode(vector, track_excess_idles=True):
	return do_encode(vector, scrambler_start_state, track_excess_idles)
#end decode

def bitstream_pretty_print(bitstream, descriptor=sys.stdout):
	descriptor.write('hbitstream = [\\\n')
	for i in range(len(bitstream)):
		sync, payload = bitstream[i]
		if i == len(bitstream) - 1:
			descriptor.write('(0x%1x, 0x%016x)]\n' % (sync, payload))
		else:
			descriptor.write('(0x%1x, 0x%016x),\\\n' % (sync, payload))
#end bitstream_pretty_print

def round(x):
	frac, intg = math.modf(x)
	if frac > 0.5:
		return long(intg)+1
	else:
		return long(intg)
#end round

def stagger_packets(packets, idle_gap, leading_idles, trailing_idles):
	# leave room for idles at the forefront
	staggered_packets = [leading_idles]
	for packet in packets:
		staggered_packets.append(packet)
		staggered_packets.append(idle_gap)
	staggered_packets[-1] = trailing_idles
	return staggered_packets
#end stagger_packets

def shuffle_packets_and_idles(packets, idles, leading_idles, trailing_idles):
	combined_vector = [leading_idles]
	while idles:
		combined_vector.append(packets[0])
		del packets[0]
		combined_vector.append(idles[0])
		del idles[0]
	assert len(packets) == 1
	combined_vector.append(packets[0])
	del packets[0]
	combined_vector.append(trailing_idles)
	return combined_vector
#end shuffle_packets_and_idles

def dilate_packets(packets, tx_rate, max_tx_rate, leading_idles, trailing_idles):
	if tx_rate >= max_tx_rate or len(packets) <= 1:
		return packets
	bytes = 0
	for packet in packets:
		bytes += len(packet)
	dilated_bytes = (1.0 * bytes * max_tx_rate) / tx_rate
	dilation_chunk = (dilated_bytes - bytes) / (len(packets) - 1)
	int_dilation_chunk = round(dilation_chunk)
	sys.stderr.write('Bytes: %d, TX rate %.2f (KB/s), dilation chunk: %.3f\n' \
		% (bytes, tx_rate / 1000.0, dilation_chunk))
	dilated_packets = [leading_idles]
	# leave room for idles at the forefront
	dilated_packets.append(0)
	for packet in packets:
		dilated_packets.append(packet)
		dilated_packets.append(round(int_dilation_chunk))
	dilated_packets[-1] = trailing_idles
	return dilated_packets
#end dilate_packets 

def main():
	idle_gap = leading_idles = trailing_idles = 0;
	packet_list_file = tx_rate = idle_list_file = None
	
	i = 1
	while i < len(sys.argv):
		if sys.argv[i] == "--constant-idle-gap":
			i += 1
			idle_gap = int(sys.argv[i])
		elif sys.argv[i] == "--list-of-idles-file":
			i += 1
			idle_list_file = sys.argv[i]
		elif sys.argv[i] == "--leading-idles":
			i += 1
			leading_idles = int(sys.argv[i])
		elif sys.argv[i] == "--trailing-idles":
			i += 1
			trailing_idles = int(sys.argv[i])
		elif sys.argv[i] == "--tx-rate":
			i += 1
			tx_rate = float(sys.argv[4]) / 8
			assert (tx_rate > 0 and tx_rate <= maxTxRate)
		else:
			packet_list_file = sys.argv[i]
		i += 1
	
	if not packet_list_file:
		sys.stderr.write("Usage %s [--tx-rate rate] [--constant-idle-gap gap] "
			"[--list-of-idles-file <file>] [--leading-idles idles] " 
			"[--trailing-idles idles] <packet list file>\n" % sys.argv[0])
		sys.exit(1)
	
	if idle_list_file and (idle_gap or tx_rate):
		sys.stderr.write("--list-of-idles-file flag supersedes "
			"--idle-gap and --tx-rate flags\n")
		idle_gap = 0
		tx_rate = 0
	
	if tx_rate and (idle_gap or leading_idles or trailing_idles):
		sys.stderr.write("--tx-rate flag supersedes --idle-gap\n")
		idle_gap = 0
	
	if idle_gap and idle_gap < INTER_FRAME_GAP:
		sys.stderr.write("Will issue at least %d idles due to inter-frame gap "
			"constraints\n" % INTER_FRAME_GAP)
		idle_gap = INTER_FRAME_GAP
	
	if (not tx_rate) and (leading_idles < 8):
		sys.stderr.write("Will issue at least 8 leading idles for scrambler sync\n")
		leading_idles = 8
	
	if (not tx_rate) and (trailing_idles < 8):
		sys.stderr.write("Will issue at least 8 trailing idles for bitstream " \
			"formatting\n")
		trailing_idles = 8
	
	packet_vector = packet_list_from_file(open(packet_list_file))
	if idle_list_file:
		gap_vector = []
		fd = open(idle_list_file)
		for line in fd:
			gap_vector.append(int(line))
		fd.close()
		if len(gap_vector) != len(packet_vector)-1:
			sys.stderr.write("Error: there should be N-1 idles, where N is "
				"the number of packets. Instead, there are %d idles and %d "
				"packets\n", len(gap_vector), packetcnt)
		packet_vector = shuffle_packets_and_idles(packet_vector, gap_vector,
			leading_idles, trailing_idles)
	elif tx_rate:
		packet_vector = dilate_packets(packet_vector, rate, maxTxRate, 
			leading_idles, trailing_idles)
	else:
		packet_vector = stagger_packets(packet_vector, idle_gap, 
			leading_idles, trailing_idles)
		
	bitstream, excess_idles = encode(packet_vector, False)
	bitstream_pretty_print(bitstream)
	sys.stderr.write('Excess idles remaining: %d\n' % excess_idles)
#end main

if __name__ == "__main__":
	main()
