
/*
 * Copyright (C) 2009 Tudor Marian <tudor.marian@cornell.edu>
 */

#include <stdint.h> /* for uint64_t */

#ifndef __SCRAMBLERCOMMON_H_
#define __SCRAMBLERCOMMON_H_

/* KiSuh added support for 64bit wide scrambling & descrambling */
static inline uint64_t scrambler64_MSB(uint64_t *state, uint64_t number)
{
	uint64_t stateval = *state, tmp, result;
	uint64_t s6 = stateval << 6;
	uint64_t s25 = stateval << 25;
	
	tmp = s6 ^ s25 ^ number;
	result = tmp ^ (tmp >> 39) ^ (tmp >> 58);

	*state = result;
	return result;	
}

static inline uint64_t descrambler64_MSB(uint64_t *state, uint64_t number)
{
	uint64_t stateval = *state, result;
	uint64_t s6 = stateval << 6;
	uint64_t s25 = stateval << 25;
	
	result = (s6 ^ s25 ^ number) ^ (number >> 39) ^ (number >> 58);
	
	*state = number;
	return result;
}

/* This should be a library call, but only some BSD flavors implement bswap64 */
static inline uint64_t bswap64(uint64_t v)
{
	/* [b0, b1, b2, b3, b4, b5, b6, b7] -> [b7, b6, b5, b4, b3, b2, b1, b0] */
	uint64_t b_0_7 = (v >> 56) | (v << 56);
	uint64_t b_1_6 = ((v & 0x00ff000000000000) >> 40) | ((v & 0x000000000000ff00) << 40);
	uint64_t b_2_5 = ((v & 0x0000ff0000000000) >> 24) | ((v & 0x0000000000ff0000) << 24);
	uint64_t b_3_4 = ((v & 0x000000ff00000000) >> 8) | ((v & 0x00000000ff000000) << 8);
	
	return (b_0_7 | b_1_6 | b_2_5 | b_3_4);
}

static inline uint64_t scrambler64_LSB(uint64_t *state, uint64_t number)
{
	uint64_t stateval = *state, tmp, result;
	uint64_t s6 = stateval >> 6;
	uint64_t s25 = stateval >> 25;
	
	tmp = s6 ^ s25 ^ bswap64(number);
	result = tmp ^ (tmp << 39) ^ (tmp << 58);
	
	*state = result;
	return bswap64(result);
}

static inline uint64_t descrambler64_LSB(uint64_t *state, uint64_t number)
{
	uint64_t stateval = *state, tmp, result;
	uint64_t s6 = stateval >> 6;
	uint64_t s25 = stateval >> 25;
	
	tmp = bswap64(number);
	result = (s6 ^ s25 ^ tmp) ^ (tmp << 39) ^ (tmp << 58);
	
	*state = tmp;
	return bswap64(result);
}

#define scrambler64 scrambler64_LSB
#define descrambler64 descrambler64_LSB

const int tap1 = (63-57);
const int tap2 = (63-38);

typedef int (*scramble_fn_ptr)(uint64_t *, int);
typedef uint64_t (*scrambler_fn)(uint64_t *, uint64_t, scramble_fn_ptr);

int scramble(uint64_t *state, int bit)
{
	uint64_t obit;
	bit = bit & 0x1;
	obit = ( ((uint64_t) bit) ^ ((*state) >> tap1) ^ ((*state) >> tap2) ) & 0x1;
	*state = ((*state) >> 1) | (obit << 63);
	return (int) obit;
}

int descramble(uint64_t *state, int bit)
{
	uint64_t obit;
	bit = bit & 0x1;
	obit = ( ((uint64_t) bit) ^ ((*state) >> tap1) ^ ((*state) >> tap2) ) & 0x1;
	*state = ((*state) >> 1) | (((uint64_t) bit) << 63);
	return (int) obit;
}

uint64_t scrambler_MSB(uint64_t *state, uint64_t number, scramble_fn_ptr fn)
{
	int i;
	uint64_t ret_number = 0x0;
	
	for (i = 63; i >= 0; i--)
	{
		int bit = (number >> i) & 0x1;
		bit = fn(state, bit);
		ret_number |= (((uint64_t) bit) << i);
	}
	return ret_number;
}

uint64_t scrambler_LSB(uint64_t *state, uint64_t number, scramble_fn_ptr fn)
{
	int i, j;
	uint64_t ret_number = 0x0;
	
	for (j = 64-8; j >= 0; j-=8)
	{
		for (i = 0; i < 8; i++)
		{
			int bit = (number >> (j+i)) & 0x1;
			bit = fn(state, bit);
			ret_number |= (((uint64_t) bit) << (j+i));
		}
	}
	return ret_number;
}

/* the default scrambler */
scrambler_fn scrambler = scrambler_LSB;

#endif /*__SCRAMBLERCOMMON_H_*/
